from pathlib import Path
from typing import List, Dict, Tuple, Literal, Any
import logging
import sys

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

sys.path.append("../src")

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import ObjectDetection
import sihl.layers

coco_labels = [
    "__background__",
    "person",
    "bicycle",
    "car",
    "motorbike",
    "aeroplane",
    "bus",
    "train",
    "truck",
    "boat",
    "trafficlight",
    "firehydrant",
    "streetsign",
    "stopsign",
    "parkingmeter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "hat",
    "backpack",
    "umbrella",
    "shoe",
    "eyeglasses",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sportsball",
    "kite",
    "baseballbat",
    "baseballglove",
    "skateboard",
    "surfboard",
    "tennisracket",
    "bottle",
    "plate",
    "wineglass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hotdog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "sofa",
    "pottedplant",
    "bed",
    "mirror",
    "diningtable",
    "window",
    "desk",
    "toilet",
    "door",
    "tvmonitor",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cellphone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "blender",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddybear",
    "hairdrier",
    "toothbrush",
    "hairbrush",
]

valid_coco_labels = coco_labels.copy()
for label in [
    "__background__",
    "streetsign",
    "hat",
    "shoe",
    "eyeglasses",
    "plate",
    "mirror",
    "window",
    "desk",
    "door",
    "blender",
    "hairbrush",
]:
    valid_coco_labels.remove(label)


def collate_fn(batch: List[Tuple[Tensor, Dict[Literal["labels", "boxes"], Any]]]):
    return (
        torch.stack([sample[0] for sample in batch]),
        {
            "classes": [
                (
                    torch.tensor(
                        [
                            valid_coco_labels.index(coco_labels[label])
                            for label in sample[1]["labels"]
                        ]
                    )
                    if "labels" in sample[1]
                    else torch.tensor([], dtype=torch.int64)
                )
                for sample in batch
            ],
            "boxes": [
                sample[1]["boxes"].data if "boxes" in sample[1] else torch.tensor([])
                for sample in batch
            ],
        },
    )

IMAGE_SIZE = 768
class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        annot_dir = self.data_dir / "annotations_trainval2017"
        self.trainset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "train2017"),
                str(annot_dir / "instances_train2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(IMAGE_SIZE - 1, max_size=IMAGE_SIZE),
                        transforms.RandomCrop(IMAGE_SIZE, pad_if_needed=True),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            )
        )
        self.valset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "val2017"),
                str(annot_dir / "instances_val2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(IMAGE_SIZE - 1, max_size=IMAGE_SIZE),
                        transforms.RandomCrop(IMAGE_SIZE, pad_if_needed=True),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            )
        )

    def train_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )


HYPERPARAMS = {
    "max_steps": 90_000,
    "batch_size": 16,
    "gradient_clip_val": 3,
    "head": {"name": "resnet50", "pretrained": True, "frozen_levels": 1},
    "neck": "FPN",
    "neck_kwargs": {"out_channels": 256, "bottom_level": 3, "top_level": 7},
    "optimizer": "SGD",
    "optimizer_kwargs": {"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
    "scheduler": "MultiStepLR",
    "scheduler_kwargs": {
        "milestones": [60_000, 80_000],
        "gamma": 0.1,
        "warmup_batches": 1000,
    },
}

if __name__ == "__main__":
    FORMAT = "%(message)s"
    logging.basicConfig(
        level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs",
        default_hp_metric=False,
        name="object_detection",
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=HYPERPARAMS["max_steps"],
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=HYPERPARAMS["gradient_clip_val"],
        precision="16-mixed",
        val_check_interval=0.25,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone(**HYPERPARAMS["head"])
        neck = getattr(sihl.layers, HYPERPARAMS["neck"])(
            backbone.out_channels, **HYPERPARAMS["neck_kwargs"]
        )
        head = ObjectDetection(
            neck.out_channels,
            num_classes=len(valid_coco_labels),
            soft_label_decay_steps=HYPERPARAMS["max_steps"],
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=getattr(torch.optim, HYPERPARAMS["optimizer"]),
            optimizer_kwargs=HYPERPARAMS["optimizer_kwargs"],
            scheduler=getattr(torch.optim.lr_scheduler, HYPERPARAMS["scheduler"]),
            scheduler_kwargs=HYPERPARAMS["scheduler_kwargs"],
            data_config={"categories": valid_coco_labels},
            hyperparameters=HYPERPARAMS,
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=HYPERPARAMS["batch_size"]))



from typing import List, Tuple, Dict, Optional
from functools import partial

from einops.layers.torch import Rearrange
from torch import nn, Tensor
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MeanMetric

from sihl.layers import (
    ConvNormAct,
    SimpleUpscaler,
    SequentialConvBlocks,
    BilinearScaler,
)
from sihl.utils import interpolate

sequential_upscalers = partial(SequentialConvBlocks, ConvBlock=SimpleUpscaler)


class Autoencoding(nn.Module):
    """Autoencoding is reconstructing an image by encoding it into a compact
    representation and then decoding that into the initial input. The point of learning
    this is that the representation can be used for downstream tasks like clustering.
    This task is self-supervised, so it can be a good choice for pre-training backbones.
    """

    def __init__(
        self,
        in_channels: List[int],
        level: int = 5,
        num_channels: int = 256,
        num_layers: int = 3,
        representation_channels: int = 1024,
        prebottleneck_size: Tuple[int, int] = (4, 4),
        activation: Optional[str] = "sigmoid",
    ):
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 1.
            representation_channels (int, optional): Number of channels in the compact representation. Defaults to 128.
            activation (Optional[str], optional): Activation function of the last layer. Defaults to "sigmoid".
        """
        assert num_channels > 0 and num_layers > 0
        assert len(in_channels) > level > 0
        super().__init__()

        self.level = level
        self.size = prebottleneck_size
        self.encoder = nn.Sequential(
            ConvNormAct(in_channels[level], num_channels, 1),
            BilinearScaler(size=self.size),
            Rearrange("b c h w -> b (c h w)", h=self.size[0], w=self.size[1]),
            nn.Linear(
                num_channels * self.size[0] * self.size[1], representation_channels
            ),
            nn.ReLU(),
        )
        self.predecoder = nn.Sequential(
            nn.Linear(
                representation_channels, num_channels * self.size[0] * self.size[1]
            ),
            nn.ReLU(),
            Rearrange("b (c h w) -> b c h w", h=self.size[0], w=self.size[1]),
        )
        self.decoder = nn.Sequential(
            sequential_upscalers(num_channels, num_channels, num_layers=level),
            SequentialConvBlocks(num_channels, num_channels, num_layers=num_layers),
            ConvNormAct(num_channels, in_channels[0], 1, norm=None, act=activation),
        )

        self.output_shapes = {
            "reconstructions": ("batch_size", in_channels[0], "height", "width"),
            "representations": ("batch_size", representation_channels),
        }

    def forward(self, inputs: List[Tensor]) -> Tensor:
        size = inputs[self.level].shape[2:]
        representations = self.encoder(inputs[self.level])
        reconstructions = interpolate(self.predecoder(representations), size=size)
        reconstructions = self.decoder(reconstructions).contiguous()
        return reconstructions, representations

    def training_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        reconstructions, representations = self.forward(inputs)
        loss = (reconstructions - targets).pow(2).mean()
        return loss, {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.mae_computer = MeanAbsoluteError()
        self.mse_computer = MeanSquaredError()

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        reconstructions, representations = self.forward(inputs)
        loss = (reconstructions - targets).pow(2).mean()
        self.loss_computer.to(loss.device).update(loss)
        self.mae_computer.to(loss.device).update(reconstructions, targets)
        self.mse_computer.to(loss.device).update(reconstructions, targets)
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "mean_absolute_error": self.mae_computer.compute().item(),
            "mean_squared_error": self.mse_computer.compute().item(),
        }



from typing import Tuple
from pathlib import Path
import logging
import sys
import json

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
from torchvision.io import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

sys.path.append("../src")

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import PanopticSegmentation
from sihl.layers import FPN


class CocoPanopticSegmentationDataset(torch.utils.data.Dataset):
    # FIXME: hardcoded "ignore" category for input padding?
    THING_CATEGORIES = [
        "person",
        "bicycle",
        "car",
        "motorcycle",
        "airplane",
        "bus",
        "train",
        "truck",
        "boat",
        "traffic light",
        "fire hydrant",
        "stop sign",
        "parking meter",
        "bench",
        "bird",
        "cat",
        "dog",
        "horse",
        "sheep",
        "cow",
        "elephant",
        "bear",
        "zebra",
        "giraffe",
        "backpack",
        "umbrella",
        "handbag",
        "tie",
        "suitcase",
        "frisbee",
        "skis",
        "snowboard",
        "sports ball",
        "kite",
        "baseball bat",
        "baseball glove",
        "skateboard",
        "surfboard",
        "tennis racket",
        "bottle",
        "wine glass",
        "cup",
        "fork",
        "knife",
        "spoon",
        "bowl",
        "banana",
        "apple",
        "sandwich",
        "orange",
        "broccoli",
        "carrot",
        "hot dog",
        "pizza",
        "donut",
        "cake",
        "chair",
        "couch",
        "potted plant",
        "bed",
        "dining table",
        "toilet",
        "tv",
        "laptop",
        "mouse",
        "remote",
        "keyboard",
        "cell phone",
        "microwave",
        "oven",
        "toaster",
        "sink",
        "refrigerator",
        "book",
        "clock",
        "vase",
        "scissors",
        "teddy bear",
        "hair drier",
        "toothbrush",
    ]
    STUFF_CATEGORIES = [
        "padding",
        "banner",
        "blanket",
        "bridge",
        "cardboard",
        "counter",
        "curtain",
        "door-stuff",
        "floor-wood",
        "flower",
        "fruit",
        "gravel",
        "house",
        "light",
        "mirror-stuff",
        "net",
        "pillow",
        "platform",
        "playingfield",
        "railroad",
        "river",
        "road",
        "roof",
        "sand",
        "sea",
        "shelf",
        "snow",
        "stairs",
        "tent",
        "towel",
        "wall-brick",
        "wall-stone",
        "wall-tile",
        "wall-wood",
        "water-other",
        "window-blind",
        "window-other",
        "tree-merged",
        "fence-merged",
        "ceiling-merged",
        "sky-other-merged",
        "cabinet-merged",
        "table-merged",
        "floor-other-merged",
        "pavement-merged",
        "mountain-merged",
        "grass-merged",
        "dirt-merged",
        "paper-merged",
        "food-other-merged",
        "building-other-merged",
        "rock-merged",
        "wall-other-merged",
        "rug-merged",
    ]
    LABEL_MAP = {
        1: "person",
        2: "bicycle",
        3: "car",
        4: "motorcycle",
        5: "airplane",
        6: "bus",
        7: "train",
        8: "truck",
        9: "boat",
        10: "traffic light",
        11: "fire hydrant",
        13: "stop sign",
        14: "parking meter",
        15: "bench",
        16: "bird",
        17: "cat",
        18: "dog",
        19: "horse",
        20: "sheep",
        21: "cow",
        22: "elephant",
        23: "bear",
        24: "zebra",
        25: "giraffe",
        27: "backpack",
        28: "umbrella",
        31: "handbag",
        32: "tie",
        33: "suitcase",
        34: "frisbee",
        35: "skis",
        36: "snowboard",
        37: "sports ball",
        38: "kite",
        39: "baseball bat",
        40: "baseball glove",
        41: "skateboard",
        42: "surfboard",
        43: "tennis racket",
        44: "bottle",
        46: "wine glass",
        47: "cup",
        48: "fork",
        49: "knife",
        50: "spoon",
        51: "bowl",
        52: "banana",
        53: "apple",
        54: "sandwich",
        55: "orange",
        56: "broccoli",
        57: "carrot",
        58: "hot dog",
        59: "pizza",
        60: "donut",
        61: "cake",
        62: "chair",
        63: "couch",
        64: "potted plant",
        65: "bed",
        67: "dining table",
        70: "toilet",
        72: "tv",
        73: "laptop",
        74: "mouse",
        75: "remote",
        76: "keyboard",
        77: "cell phone",
        78: "microwave",
        79: "oven",
        80: "toaster",
        81: "sink",
        82: "refrigerator",
        84: "book",
        85: "clock",
        86: "vase",
        87: "scissors",
        88: "teddy bear",
        89: "hair drier",
        90: "toothbrush",
        92: "banner",
        93: "blanket",
        95: "bridge",
        100: "cardboard",
        107: "counter",
        109: "curtain",
        112: "door-stuff",
        118: "floor-wood",
        119: "flower",
        122: "fruit",
        125: "gravel",
        128: "house",
        130: "light",
        133: "mirror-stuff",
        138: "net",
        141: "pillow",
        144: "platform",
        145: "playingfield",
        147: "railroad",
        148: "river",
        149: "road",
        151: "roof",
        154: "sand",
        155: "sea",
        156: "shelf",
        159: "snow",
        161: "stairs",
        166: "tent",
        168: "towel",
        171: "wall-brick",
        175: "wall-stone",
        176: "wall-tile",
        177: "wall-wood",
        178: "water-other",
        180: "window-blind",
        181: "window-other",
        184: "tree-merged",
        185: "fence-merged",
        186: "ceiling-merged",
        187: "sky-other-merged",
        188: "cabinet-merged",
        189: "table-merged",
        190: "floor-other-merged",
        191: "pavement-merged",
        192: "mountain-merged",
        193: "grass-merged",
        194: "dirt-merged",
        195: "paper-merged",
        196: "food-other-merged",
        197: "building-other-merged",
        198: "rock-merged",
        199: "wall-other-merged",
        200: "rug-merged",
    }
    ALL_CATEGORIES = STUFF_CATEGORIES + THING_CATEGORIES
    PAD = STUFF_CATEGORIES.index("padding")
    IGNORE = len(ALL_CATEGORIES)

    def __init__(self, data_dir: Path, train: bool = False) -> None:
        self.image_size = 640  # 800
        self.train = train
        self.data_dir = data_dir
        self.annots_by_image = {}
        split = "train" if train else "val"
        self.image_dir = self.data_dir / f"{split}2017"
        self.annot_dir = (
            self.data_dir
            / "panoptic_annotations_trainval2017"
            / f"panoptic_{split}2017"
        )

        with open(
            self.data_dir
            / "panoptic_annotations_trainval2017"
            / f"panoptic_{split}2017.json"
        ) as f:
            coco_data = json.load(f)

        image_by_id = {
            image_annot["id"]: image_annot["file_name"]
            for image_annot in coco_data["images"]
        }
        for annot in coco_data["annotations"]:
            image_name = image_by_id[annot["image_id"]]
            if (
                not (self.image_dir / image_name).exists()
                or not (self.annot_dir / annot["file_name"]).exists()
            ):
                continue
            image_path = str((self.image_dir / image_name).resolve())
            self.annots_by_image[image_path] = annot
        self.annots_by_image = [
            (key, value) for key, value in self.annots_by_image.items()
        ]

        if self.train:
            self.transform = transforms.Compose(
                [
                    transforms.Resize(self.image_size - 1, max_size=self.image_size),
                    transforms.RandomCrop(
                        self.image_size,
                        pad_if_needed=True,
                        fill={tv_tensors.Image: 0, tv_tensors.Mask: self.PAD},
                    ),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.Resize(self.image_size - 1, max_size=self.image_size),
                    transforms.RandomCrop(
                        self.image_size,
                        pad_if_needed=True,
                        fill={tv_tensors.Image: 0, tv_tensors.Mask: self.PAD},
                    ),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, annot = self.annots_by_image[idx]
        image = torchvision.io.read_image(image_path, mode=ImageReadMode.RGB)
        target = torchvision.io.read_image(
            str((self.annot_dir / annot["file_name"]).resolve()), mode=ImageReadMode.RGB
        )
        r, g, b = target.to(torch.int64)
        id_map = r + 256 * g + 256**2 * b
        semantic_map = id_map.clone()
        semantic_map[semantic_map == 0] = self.IGNORE
        for segment in annot["segments_info"]:
            category = self.ALL_CATEGORIES.index(self.LABEL_MAP[segment["category_id"]])
            semantic_map[semantic_map == segment["id"]] = category
        target = torch.stack([semantic_map, id_map])
        image, target = self.transform(image, tv_tensors.Mask(target))
        return image, target.to(torch.int64)

    def __len__(self) -> int:
        return len(self.annots_by_image)


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"
        self.data_dir.mkdir(exist_ok=True)

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = CocoPanopticSegmentationDataset(self.data_dir, train=True)
        self.validset = CocoPanopticSegmentationDataset(self.data_dir, train=False)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="panoptic_segmentation"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=90_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        precision="16-mixed",
        val_check_interval=0.25,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True, frozen_levels=1)
        neck = FPN(backbone.out_channels, out_channels=256, bottom_level=3, top_level=7)
        head = PanopticSegmentation(
            in_channels=neck.out_channels,
            num_stuff_classes=len(CocoPanopticSegmentationDataset.STUFF_CATEGORIES),
            num_thing_classes=len(CocoPanopticSegmentationDataset.THING_CATEGORIES),
            mask_top_level=5,
            soft_label_decay_steps=90_000,
            ignore_index=CocoPanopticSegmentationDataset.IGNORE,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-4},
            data_config={
                "categories": CocoPanopticSegmentationDataset.ALL_CATEGORIES,
                "ignore": CocoPanopticSegmentationDataset.IGNORE,
            },
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))



from pathlib import Path
from typing import Tuple, List
import logging

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision.io.image import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule
from sihl.heads import MulticlassClassification
from sihl.torchvision_backbone import TorchvisionBackbone


class StanfordCarsDataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, train: bool) -> None:
        self.data_dir = data_dir
        self.images: List[Tuple[Path, int]] = []
        if train:
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

        img_dir = data_dir / "car_data" / "car_data" / ("train" if train else "test")
        self.categories = sorted([_.name for _ in img_dir.iterdir()])
        for img in img_dir.rglob("*.jpg"):
            annot = img.parent.name
            self.images.append((img, self.categories.index(annot)))

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target = self.images[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), mode=ImageReadMode.RGB
        )
        return self.transform(image), torch.tensor(target)

    def __len__(self) -> int:
        return len(self.images)


class StanfordCarsDataModule(pl.LightningDataModule):
    def __init__(self, batch_size) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "StanfordCars"

    def prepare_data(self) -> None:
        print(self.data_dir)
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "jutrera/stanford-car-dataset-by-classes-folder",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = StanfordCarsDataset(self.data_dir, train=True)
        self.validset = StanfordCarsDataset(self.data_dir, train=False)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            drop_last=True,
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            drop_last=True,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="multiclass_classification"
    )
    lightning.seed_everything(0)
    datamodule = StanfordCarsDataModule(batch_size=64)
    datamodule.prepare_data()
    datamodule.setup()

    EPOCHS = 50
    trainer = pl.Trainer(
        max_epochs=EPOCHS,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = MulticlassClassification(
            in_channels=backbone.out_channels,
            num_classes=len(datamodule.trainset.categories),
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
            scheduler=torch.optim.lr_scheduler.OneCycleLR,
            scheduler_kwargs={"max_lr": 1e-3, "epochs": EPOCHS, "steps_per_epoch": 127},
            data_config={"categories": datamodule.trainset.categories},
        )

    log.info(
        summary(
            model.model,
            input_data=model.model.backbone.dummy_input,
            row_settings=("var_names",),
            col_names=("output_size", "num_params", "mult_adds", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=datamodule)



from typing import List

from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import Autoencoding

from .common import get_images, plot_to_numpy


@get_images.register(Autoencoding)
def _(head, config, input, target, features) -> List[np.ndarray]:
    reconstructions, embeddings = head(features)
    reconstructions = (
        (reconstructions.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    )
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    target = (target.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)
        axes[1].title.set_text("Target")
        axes[1].imshow(target[batch])
        axes[2].title.set_text("Prediction")
        axes[2].imshow(reconstructions[batch])
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import List

from torch import nn, Tensor
import torch

from sihl.layers.convblocks import ConvNormAct
from sihl.layers.scalers import BilinearScaler, AntialiasedDownscaler
from sihl.utils import EPS


class FastNormalizedFusion(nn.Module):
    def __init__(self, num_inputs: int = 2) -> None:
        super().__init__()
        self.num_inputs = num_inputs
        self.weights = nn.Parameter(torch.ones(num_inputs), requires_grad=True)

    def forward(self, inputs: List[Tensor]) -> Tensor:
        weights = self.weights.relu().to(inputs[0].device)
        weights = weights / (weights.sum() + EPS)
        return torch.stack(
            [weights[idx] * inputs[idx] for idx in range(self.num_inputs)]
        ).sum(dim=0)


class BiFPNLayer(nn.Module):
    def __init__(self, out_channels: int, num_levels: int) -> None:
        super().__init__()
        levels = range(num_levels - 1)
        self.upscalers = nn.ModuleList(BilinearScaler(scale=2) for level in levels)
        self.up_fusions = nn.ModuleList(FastNormalizedFusion(2) for level in levels)
        self.up_convs = nn.ModuleList(
            ConvNormAct(out_channels, out_channels) for level in levels
        )
        self.downscalers = nn.ModuleList(
            AntialiasedDownscaler(out_channels, out_channels) for level in levels
        )
        self.down_fusions = nn.ModuleList(FastNormalizedFusion(3) for level in levels)
        self.down_fusions[-1] = FastNormalizedFusion(2)
        self.down_convs = nn.ModuleList(
            [ConvNormAct(out_channels, out_channels) for level in levels]
        )

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        top_down = [inputs[-1]]
        for idx, (conv, fusion, upscaler) in enumerate(
            zip(self.up_convs, self.up_fusions, self.upscalers)
        ):
            top_down.append(conv(fusion([upscaler(top_down[-1]), inputs[-idx - 2]])))

        bottom_up = [top_down[-1]]
        for idx, (conv, fusion, downscaler) in enumerate(
            zip(self.down_convs, self.down_fusions, self.downscalers)
        ):
            args = [downscaler(bottom_up[-1]), inputs[idx + 1]]
            if idx != len(inputs) - 1:
                args = args + [top_down[-idx - 2]]
            bottom_up.append(conv(fusion(args)))
        return bottom_up


class BiFPN(nn.Module):
    """https://arxiv.org/abs/1911.09070"""

    def __init__(
        self,
        in_channels: List[int],
        out_channels: int,
        bottom_level: int,
        top_level: int,
        num_layers: int = 3,
    ):
        super().__init__()
        assert num_layers > 0
        assert 0 < bottom_level < top_level
        self.out_channels = in_channels[:bottom_level] + [
            out_channels for level in range(bottom_level, top_level + 1)
        ]
        self.bottom_level = bottom_level
        self.top_level = top_level
        self.lateral_connections = nn.ModuleList(
            ConvNormAct(in_c, out_channels, kernel_size=1)
            for in_c in in_channels[bottom_level : top_level + 1]
        )
        self.downscalers = nn.ModuleList(
            AntialiasedDownscaler(out_channels, out_channels)
            for _ in range(top_level + 1 - len(in_channels))
        )
        num_levels = top_level - bottom_level + 1
        self.layers = nn.Sequential(
            *(BiFPNLayer(out_channels, num_levels) for layer in range(num_layers))
        )

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        features = [
            lateral(inputs[self.bottom_level + idx])
            for idx, lateral in enumerate(self.lateral_connections)
        ]
        for downscaler in self.downscalers:
            features.append(downscaler(features[-1]))
        outs = self.layers(features)
        return inputs[: self.bottom_level] + outs + inputs[self.top_level + 1 :]



from copy import deepcopy
from typing import Any
from typing import Tuple, List, Dict, Union, Type, Optional
import warnings

from lightning.fabric.utilities.exceptions import MisconfigurationException
from torch import nn, Tensor
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
import lightning.pytorch as pl
import torch

from sihl.heads import Head, ViewInvarianceLearning
from sihl.sihl_model import SihlModel
from sihl.visualization import visualize

# https://github.com/Lightning-AI/pytorch-lightning/issues/5558#issuecomment-1199306489
warnings.filterwarnings("ignore", "Detected call of", UserWarning)
torch.set_float32_matmul_precision("high")


class SihlLightningModule(pl.LightningModule):
    def __init__(
        self,
        model: SihlModel,
        optimizer: Type[Optimizer] = torch.optim.Adam,
        optimizer_kwargs: Optional[Dict[str, Any]] = None,
        scheduler: Optional[Type[LRScheduler]] = None,
        scheduler_kwargs: Optional[Dict[str, Any]] = None,
        data_config: Optional[List[Dict[str, Any]]] = None,
        hyperparameters: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.optimizer_kwargs = optimizer_kwargs or {}
        self.scheduler = scheduler
        self.scheduler_kwargs = scheduler_kwargs or {}
        self.data_config = data_config or [{} for _ in range(len(self.model.heads))]
        if isinstance(self.data_config, dict):
            self.data_config = [self.data_config]
        self.visualization_batches = []
        self.visualize_n_batches = 5
        self.visualize_n_per_batch = 1
        self.hyperparameters = hyperparameters
        self.log_kwargs = dict(on_epoch=False, on_step=True, prog_bar=True)

    def forward(self, x: Tensor) -> List[Union[Tensor, List[Tensor]]]:
        return self.model(x)

    def on_fit_start(self):
        self = self.to(memory_format=torch.channels_last)

    def on_train_epoch_start(self) -> None:
        for head in self.model.heads:
            head: Head
            if hasattr(head, "on_train_epoch_start"):
                head.on_train_epoch_start()

    def on_train_epoch_end(self) -> None:
        for head in self.model.heads:
            head: Head
            if hasattr(head, "on_train_epoch_end"):
                head.on_train_epoch_end()

    def training_step(
        self, batch: Union[Tuple[Tensor, Any], Tensor], batch_idx: int
    ) -> Tensor:
        x, targets = (batch, None) if isinstance(batch, Tensor) else batch
        # FIXME: heads' registered buffers aren't moved to device by default
        self.to(x.device)
        if not isinstance(targets, list):
            targets = [targets]  # single-headed

        head_inputs = self.model.extract_features(x)
        losses = []
        for head_idx, (head, target) in enumerate(zip(self.model.heads, targets)):
            head: Head
            if isinstance(head, ViewInvarianceLearning):  # FIXME: this is a hack
                target = self.model.extract_features(target)  # second view of inputs

            if isinstance(target, dict):
                loss, metrics = head.training_step(head_inputs, **target)
            else:
                loss, metrics = head.training_step(head_inputs, target)

            metrics = {f"head{head_idx}/train/{k}": v for k, v in metrics.items()}
            try:
                self.log(f"head{head_idx}/train/loss", loss, **self.log_kwargs)
                self.log_dict(metrics, **self.log_kwargs)
            except MisconfigurationException:
                pass
            losses.append(loss)
        loss = torch.stack(losses).sum()
        scheduler = self.lr_schedulers()
        try:
            if isinstance(scheduler, LRScheduler):
                self.log("learning_rate", scheduler.get_last_lr()[0], **self.log_kwargs)
            else:
                lightning_optimizer = self.optimizers()
                for param_group in lightning_optimizer.optimizer.param_groups:
                    self.log("learning_rate", param_group["lr"], **self.log_kwargs)
        except (MisconfigurationException, AttributeError):
            pass
        torch.cuda.empty_cache()
        return loss

    def validation_step(self, batch: Tuple[Tensor, Any], batch_idx: int) -> Tensor:
        x, targets = (batch, None) if isinstance(batch, Tensor) else batch
        if not isinstance(targets, list):
            targets = [targets]

        if len(self.visualization_batches) < self.visualize_n_batches:
            batch_idxs = [_["idx"] for _ in self.visualization_batches]
            if batch_idx not in batch_idxs:
                self.visualization_batches.append(
                    {
                        "input": x[: self.visualize_n_per_batch],
                        "targets": targets,
                        "idx": batch_idx,
                    }
                )

        head_inputs = self.model.extract_features(x)
        total_loss = 0
        for head_idx, (head, target) in enumerate(zip(self.model.heads, targets)):
            head: Head
            if isinstance(head, ViewInvarianceLearning):  # FIXME: this is a hack
                target = self.model.extract_features(target)

            if isinstance(target, dict):
                head_loss, metrics = head.validation_step(head_inputs, **target)
            else:
                head_loss, metrics = head.validation_step(head_inputs, target)

            total_loss = head_loss + total_loss
            metrics = {f"head{head_idx}/valid/{k}": v for k, v in metrics.items()}
            try:
                self.log_dict(metrics, on_epoch=True, on_step=False, prog_bar=True)
            except MisconfigurationException:
                pass
        torch.cuda.empty_cache()
        return total_loss

    def pretraining_step(self, batch: Tuple[Tensor, Any]) -> None:
        x, targets = (batch, None) if isinstance(batch, Tensor) else batch
        if not isinstance(targets, list):
            targets = [targets]

        head_inputs = self.model.extract_features(x.to("cuda"))  # FIXME
        for head_idx, (head, target) in enumerate(zip(self.model.heads, targets)):
            head: Head
            if not hasattr(head, "pretraining_step"):
                continue
            if isinstance(head, ViewInvarianceLearning):  # FIXME: this is a hack
                target = self.model.extract_features(target)

            if isinstance(target, dict):
                head.pretraining_step(head_inputs, **target)
            else:
                head.pretraining_step(head_inputs, target)

        torch.cuda.empty_cache()

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[List[Optimizer], List[LRScheduler]]]:
        optimizer_kwargs = deepcopy(self.optimizer_kwargs)

        # It can be beneficial to use a lower lr for the backbone when finetuning
        base_lr, backbone_lr_factor = optimizer_kwargs.get("lr", 1e-3), 1.0
        if "backbone_lr_factor" in optimizer_kwargs:
            backbone_lr_factor = optimizer_kwargs.pop("backbone_lr_factor")

        # backbone_module_names = {mn for mn, m in self.model.backbone.named_modules()}
        backbone_params = set(self.model.backbone.parameters())
        non_backbone_params = set(self.parameters()) - backbone_params
        parameters = [
            {"params": list(non_backbone_params)},
            {"params": list(backbone_params), "lr": base_lr * backbone_lr_factor},
        ]

        if "weight_decay" in optimizer_kwargs:
            # Certain layers should not be subject to weight decay, so we isolate them
            blacklist = (nn.LayerNorm, nn.GroupNorm, nn.BatchNorm2d, nn.Embedding)
            no_decay = set()
            for mn, m in self.named_modules():
                for pn, p in m.named_parameters():
                    if pn.endswith("bias") or isinstance(m, blacklist):
                        no_decay.add(p)
            decay = set(self.parameters()) - no_decay
            parameters = [
                {"params": list(non_backbone_params & decay)},
                {"params": list(non_backbone_params & no_decay), "weight_decay": 0.0},
                {
                    "params": list(backbone_params & decay),
                    "lr": base_lr * backbone_lr_factor,
                },
                {
                    "params": list(backbone_params & no_decay),
                    "lr": base_lr * backbone_lr_factor,
                    "weight_decay": 0.0,
                },
            ]

        optimizer = self.optimizer(parameters, **optimizer_kwargs)

        if self.scheduler:
            warmup_batches = None
            if "warmup_batches" in self.scheduler_kwargs:
                warmup_batches = self.scheduler_kwargs.pop("warmup_batches")
            scheduler = self.scheduler(optimizer, **self.scheduler_kwargs)
            if warmup_batches:
                scheduler = torch.optim.lr_scheduler.SequentialLR(
                    optimizer,
                    [
                        torch.optim.lr_scheduler.LinearLR(
                            optimizer, start_factor=0.01, total_iters=warmup_batches
                        ),
                        scheduler,
                    ],
                    milestones=[warmup_batches],
                )
            scheduler_config = {"scheduler": scheduler, "interval": "step"}
            scheduler_config["monitor"] = "loss"
            return {"optimizer": optimizer, "lr_scheduler": scheduler_config}
        return optimizer

    def on_validation_start(self) -> None:
        for head in self.model.heads:
            head: Head
            if hasattr(head, "on_validation_start"):
                head.on_validation_start()

    def on_validation_epoch_end(self) -> None:
        for head_idx, head in enumerate(self.model.heads):
            head: Head
            if not hasattr(head, "on_validation_end"):
                continue
            metrics = {
                f"head{head_idx}/valid/{k}": v
                for k, v in head.on_validation_end().items()
            }
            try:
                self.log_dict(metrics, sync_dist=True)
                if self.hyperparameters:
                    for logger in self.loggers:
                        try:
                            logger.log_hyperparams(
                                self.hyperparameters, metrics, step=self.global_step
                            )
                        except BaseException:
                            pass
            except MisconfigurationException:
                pass

        for viz_batch_idx, viz_batch in enumerate(self.visualization_batches):
            visualize(
                model=self.model,
                configs=self.data_config,
                input=viz_batch["input"],
                targets=viz_batch["targets"],
                logger=self.logger,
                step=self.global_step,
                start_idx=viz_batch_idx * self.visualize_n_per_batch,
            )

        torch.cuda.empty_cache()

    def on_pretraining_start(self) -> None:
        for head in self.model.heads:
            head: Head
            if hasattr(head, "on_pretraining_start"):
                head.on_pretraining_start()

    def on_pretraining_end(self) -> None:
        for head in self.model.heads:
            head: Head
            if hasattr(head, "on_pretraining_end"):
                head.on_pretraining_end()



from pathlib import Path
from typing import Tuple, Literal

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision.io.image import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import logging
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule
from sihl.heads import Autoencoding
from sihl.torchvision_backbone import TorchvisionBackbone


class StanfordCarsDataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, split: Literal["train", "test"]) -> None:
        self.split = split
        self.data_dir = data_dir
        self.images = list((data_dir / "car_data" / "car_data" / split).rglob("*.jpg"))

        if split == "train":
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path = self.images[idx]
        image = self.transform(
            torchvision.io.read_image(str(image_path.resolve()), mode=ImageReadMode.RGB)
        )
        return image, image

    def __len__(self) -> int:
        return len(self.images)


class StanfordCarsDataModule(pl.LightningDataModule):
    def __init__(self, batch_size) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "StanfordCars"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "jutrera/stanford-car-dataset-by-classes-folder",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = StanfordCarsDataset(self.data_dir, split="train")
        self.validset = StanfordCarsDataset(self.data_dir, split="test")

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="autoencoding"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_epochs=200,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = Autoencoding(
            in_channels=backbone.out_channels,
            prebottleneck_size=(8, 8),
            representation_channels=1024,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
        )

    log.debug(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=StanfordCarsDataModule(batch_size=32))
